"""
This module contains a convertor from HdlConvertor AST
to a BasicHldSimModel code.

BasicHldSimModel is a simple HDL like python component description
which can be directly simulated by Discrete-event simulators like
https://github.com/Nic30/hwtSimApi.git
"""

from hdlConvertorAst.to.basic_hdl_sim_model._main import ToBasicHdlSimModel
